(** @author: Benoît Montagu <benoit.montagu@inria.fr>
    Copyright © Inria 2020-2021
*)

open Lang.Ast

module Make
    (Log: Logs.S)
    (Options: sig
       val debug: bool
     end)
    (Calls: Calls.S)
    (CtxtEnv: sig
       type t
       val empty: t
       val compare: t -> t -> int
       val pp: Format.formatter -> t -> unit
     end)
    (Abs: sig
       type t
       val bot: t
       val leq: t -> t -> bool
       val join: t -> t -> t
       val widen: t -> t -> t
       val pp: Format.formatter -> t -> unit
     end)
= struct
  (** Global environments are maps whose keys are pairs made of a
     variable and a calling context (i.e. addresses, in the AAM
     terminology). The values that global environments contain are the
     abstract values of the module [Abs]. *)
  module GEnv = struct
    (** Variables with their calling context *)
    module V = struct
      type t = PP.t * Calls.t
      [@@deriving ord]

      let pp fmt (x, ctxt) =
        Format.fprintf fmt "%a @@ [%a]"
          PP.pp x
          Calls.pp ctxt
    end

    module M = MapMake(V)

    type key = V.t
    type value = Abs.t
    type t = value M.t

    let singleton = M.singleton

    let get k m =
      match M.find_opt k m with
      | None -> Abs.bot
      | Some a -> a

    let bot = M.empty

    let leq c1 c2 =
      M.for_all (fun k v1 ->
          match M.find_opt k c2 with
          | None -> false
          | Some v2 -> Abs.leq v1 v2
        )
        c1

    let join c1 c2 =
      M.union (fun _k v1 v2 -> Some (Abs.join v1 v2)) c1 c2

    let widen c1 c2 =
      M.union (fun _k v1 v2 -> Some (Abs.widen v1 v2)) c1 c2

    let pp fmt c =
      Format.fprintf fmt "@[{%a}@]"
        (Format.pp_print_list
           ~pp_sep:(fun fmt () -> Format.fprintf fmt ";@ ")
           (fun fmt (k, v) ->
              Format.fprintf fmt "@[%a@ ->@ %a@]"
                V.pp k
                Abs.pp v
           )
        )
        (M.bindings c)

    let find_opt = M.find_opt
    let union = M.union
    let merge = M.merge
    let iter = M.iter
    let cardinal = M.cardinal
  end

  module BoolGEnv = struct
    type key = GEnv.key
    type value = bool
    type t = value GEnv.M.t
    let empty = GEnv.M.empty
    let find_opt = GEnv.M.find_opt
    let add = GEnv.M.add
  end

  (** Pairs of a global environment and an abstract value: such pairs
     are the results of the analyzer. *)
  module GEnvAbs = struct
    type t = GEnv.t * Abs.t

    let bot = (GEnv.bot, Abs.bot)

    let leq (genv1, a1) (genv2, a2) =
      GEnv.leq genv1 genv2 && Abs.leq a1 a2

    let join (genv1, a1) (genv2, a2) =
      (GEnv.join genv1 genv2, Abs.join a1 a2)

    let widen (genv1, a1) (genv2, a2) =
      (GEnv.widen genv1 genv2, Abs.widen a1 a2)

    let pp fmt (genv, a) =
      Format.fprintf fmt "@[{ R =@ %a;@ V =@ %a@ }@]"
        GEnv.pp genv
        Abs.pp a
  end

  (** Memoization module for the analyzer *)
  module MemoFix = struct
    module CallsTermEnv = struct
      type t = Calls.t * Term.t * CtxtEnv.t
      [@@deriving ord]

      let pp fmt (calls, t, env) =
        Format.fprintf fmt "@[@[[%a]@]@ %a@ %a@]@."
          Calls.pp calls
          Term.pp t
          CtxtEnv.pp env
    end

    module Unit = struct
      type t = unit

      let compare () () = 0
      let pp _fmt () = ()
    end

    module AbsClosure = struct
      type t = (Calls.t * Term.t * CtxtEnv.t) * unit
      [@@deriving ord]

      let pp fmt ((calls, t, env), ()) =
        Format.fprintf fmt "@[@[[%a]@]@ %a@ %a@]@."
          Calls.pp calls
          Term.pp t
          CtxtEnv.pp env

      let widen (((_c1, _t1, _e1), ()) as a1) ((_c2, _t2, _e2), ()) = a1
      (* there should be no widening at all *)
    end

    include
      Memo.MakeFix(Log)(AbsClosure)(Map2.Make(CallsTermEnv)(Unit))(GEnvAbs)
  end

  (** Helper function to define the abstract evaluator *)
  let cofix =
    let curry f calls env t = f ((calls, t, env), ())
    and curry1 f b calls env t = f b ((calls, t, env), ()) in
    let uncurry2 f g b ((calls, t, env), ()) = f (curry1 g) b calls env t in
    fun f ->
      curry @@
      MemoFix.cofix ~debug:Options.debug ~transient:false @@
      uncurry2 f

  (** Outer fixpoint computation *)
  let fix leq join bot f =
    let rec loop ((n, x) as acc) =
      if Options.debug
      then Log.eprintf "Outer iteration %i:@." n;
      let y = f x in
      if leq y x
      then acc
      else loop (n+1, join x y)
    in
    let (n, y) = loop (1, bot) in
    if Options.debug
    then Log.eprintf "Total outer iterations: %i@." n;
    (n, y)

  (** The final evaluator: computes a fixpoint of the inner evaluator *)
  let make_eval ~print_stats ~check_genv eval t =
    let stats = ref Memo.Statistics.zero in
    let (nb_iter, (genv, v)) =
      fix GEnvAbs.leq GEnvAbs.widen GEnvAbs.bot
        (fun (genv, _v) ->
           let (res, stats') = eval genv Calls.empty CtxtEnv.empty t in
           stats := stats';
           res
        )
    in
    stats := { !stats with iterations = nb_iter-1 };
    print_stats genv;
    check_genv genv;
    (v, !stats)

end
