(** @author: Benoît Montagu <benoit.montagu@inria.fr>
    Copyright © Inria 2020-2021
*)

open Lang.Ast

(** The abstract domain for values, parameterized by the contexts for
   closures, the domain for booleans and the domain for integers. *)
module AbsDomain
    (CtxtEnv: sig
       type t
       val leq: t -> t -> bool
       val join: t -> t -> t
       val meet: t -> t -> t
       val pp: Format.formatter -> t -> unit
     end)
    (Bools: AbstractDomains.Sigs.BOOL_DOMAIN)
    (Ints: AbstractDomains.Sigs.INT_DOMAIN) : sig
  type t = { bools: Bools.t; ints: Ints.t; closures: closures }
  and[@unboxed] closures =
    | Clos of CtxtEnv.t Map.Make(Fun).t
  val bot: t
  val is_bot: t -> bool
  val singleton: [< `Bool of bool | `Int of int | `Closure of Fun.t * CtxtEnv.t] -> t
  val ints: Ints.t -> t
  val bools: Bools.t -> t
  val mem: [< `Bool of bool | `Int of int | `Clos of Fun.t] -> t -> bool
  val leq: t -> t -> bool
  val join: t -> t -> t
  val closures_join_map:
    join:('a -> 'a -> 'a) -> bot:'a ->
    (Fun.t -> CtxtEnv.t -> 'a) -> t -> 'a
  val closures_iter: (Fun.t -> CtxtEnv.t -> unit) -> t -> unit
  val widen: t -> t -> t
  val meet: t -> t -> t
  val pp: Format.formatter -> t -> unit
end = struct
  module M = Map.Make(Fun)

  (** The possible booleans, integers, and closures *)
  type t = { bools: Bools.t; ints: Ints.t; closures: closures }

  (** A closure is a map that associates to every possible function an
     abstract context *)
  and[@unboxed] closures =
    | Clos of CtxtEnv.t M.t

  let clos_bot = Clos M.empty

  let bot = { bools = Bools.bot; ints = Ints.bot; closures = clos_bot }

  let clos_is_bot = function
    | Clos m -> M.is_empty m

  let is_bot { bools; ints; closures } =
    Bools.is_bot bools &&
    Ints.is_bot ints &&
    clos_is_bot closures

  let clos_is_top = function
    | Clos _ -> false

  let is_top { bools; ints; closures } =
    Bools.is_top bools &&
    Ints.is_top ints &&
    clos_is_top closures

  let bool_singleton b =
    { bools = Bools.singleton b;
      ints = Ints.bot;
      closures = clos_bot;
    }

  let int_singleton n =
    { bools = Bools.bot;
      ints = Ints.singleton n;
      closures = clos_bot;
    }

  let closure_singleton t env =
    { bools = Bools.bot;
      ints = Ints.bot;
      closures = Clos (M.singleton t env);
    }

  let singleton = function
    | `Bool b -> bool_singleton b
    | `Int n -> int_singleton n
    | `Closure (t, env) -> closure_singleton t env

  let ints i =
    { bools = Bools.bot;
      ints = i;
      closures = clos_bot;
    }

  let bools b =
    { bools = b;
      ints = Ints.bot;
      closures = clos_bot;
    }

  let rec join
      { bools = bools1; ints = ints1; closures = clos1 }
      { bools = bools2; ints = ints2; closures = clos2 } =
    { bools = Bools.join bools1 bools2;
      ints = Ints.join ints1 ints2;
      closures = clos_join clos1 clos2;
    }

  and clos_join c1 c2 = match c1, c2 with
    | Clos m1, Clos m2 ->
      let m =
        M.union (fun _t envs1 envs2 ->
                   Some (CtxtEnv.join envs1 envs2)
          )
          m1 m2
      in
      Clos m

  let rec meet
      { bools = bools1; ints = ints1; closures = clos1 }
      { bools = bools2; ints = ints2; closures = clos2 } =
    { bools = Bools.meet bools1 bools2;
      ints = Ints.meet ints1 ints2;
      closures = clos_meet clos1 clos2;
    }

  and clos_meet c1 c2 = match c1, c2 with
    | Clos m1, Clos m2 ->
      let m =
        M.union (fun _t envs1 envs2 ->
                   Some (CtxtEnv.meet envs1 envs2)
          )
          m1 m2
      in
      Clos m

  let rec leq
      { bools = bools1; ints = ints1; closures = clos1 }
      { bools = bools2; ints = ints2; closures = clos2 }
    =
    Bools.leq bools1 bools2 &&
    Ints.leq ints1 ints2 &&
    clos_leq clos1 clos2

  and clos_leq c1 c2 = match c1, c2 with
    | Clos m1, Clos m2 ->
      M.for_all (fun t envs1 ->
          match M.find_opt t m2 with
          | None -> false
          | Some envs2 -> CtxtEnv.leq envs1 envs2
        )
        m1

  let clos_mem t = function
    | Clos m -> M.mem t m

  let mem absval { bools; ints; closures } =
    match absval with
    | `Bool b -> Bools.mem b bools
    | `Int n -> Ints.mem n ints
    | `Clos t -> clos_mem t closures

  let closures_join_map ~join ~bot fclos r = match r.closures with
    | Clos m ->
      M.fold (fun t c acc -> join (fclos t c) acc) m bot

  let closures_iter f r = match r.closures with
    | Clos m -> M.iter f m

  let widen
      { bools = bools1; ints = ints1; closures = clos1 }
      { bools = bools2; ints = ints2; closures = clos2 } =
    { bools = Bools.widen bools1 bools2;
      ints = Ints.widen ints1 ints2;
      closures = clos_join clos1 clos2;
      (* no widening on closures: it is a finite-height domain *)
    }

  let rec pp fmt ({ bools; ints; closures } as v) =
    if is_bot v
    then Format.fprintf fmt "⊥"
    else if is_top v
    then Format.fprintf fmt "⊤"
    else
      Format.fprintf fmt
        "{ @[@[bools =@ %a@];@ @[ints =@ %a@];@ @[closures =@ %a@]@] }"
        Bools.pp bools
        Ints.pp ints
        pp_closures closures

  and pp_closures fmt = function
    | Clos m ->
      if M.is_empty m
      then Format.pp_print_string fmt "∅"
      else
        Format.fprintf fmt "@[{%a}@]"
          (Format.pp_print_list
             ~pp_sep:(fun fmt () -> Format.fprintf fmt ",@ ")
             (fun fmt (x, mx) ->
                Format.fprintf fmt "@[%a@ %a@]"
                  (Fun.pp_rec true) x
                  CtxtEnv.pp mx
             )
          )
          (M.bindings m)
end

(** Main analyzer module, parameterized by logging functions, options,
   the domain for calls, the domain for booleans, and the domain for
   integers. *)
module Analyze
    (Log: Logs.S)
    (Options: sig
       val refine_branches: bool
       val branch_is_context: bool
       val let_is_context: bool
       val no_env_restrict: bool
       val kill_unreachable: bool
       val debug: bool
     end)
    (Calls: Calls.S)
    (Bools: AbstractDomains.Sigs.BOOL_DOMAIN)
    (Ints: AbstractDomains.Sigs.INT_DOMAIN) =
struct

  (** Sets of pairs of a program point and a calling context *)
  module CallsSet = struct
    module C = struct
      type t = PP.t * Calls.t
      [@@deriving ord]

      let pp fmt (pp, calls) =
        Format.fprintf fmt "%a @@ %a"
          PP.pp pp
          Calls.pp calls
    end
    module S = Set.Make(C)

    type t = S.t

    let compare = S.compare

    let singleton = S.singleton

    let leq = S.subset

    let iter = S.iter

    let join = S.union

    let meet = S.inter

    let join_map join map s bot =
      S.fold (fun x acc -> join acc (map x)) s bot

    let pp fmt s =
      let l = S.elements s in
      let fmt_string = match l with
        | [_] -> ("@[%a@]" : _ format)
        | _ -> "@[{%a}@]"
      in
      Format.fprintf fmt fmt_string
        (Format.pp_print_list
           ~pp_sep:(fun fmt () -> Format.fprintf fmt ",@ ")
           C.pp
        )
        l
  end

  (** Abstract contexts, i.e. environments that map variables to sets
     of pairs of a program point and a calling context. *)
  module CtxtEnv = struct
    type t = CallsSet.t Env.t
    [@@deriving ord]

    let empty = Env.empty

    let leq env1 env2 =
      Env.equal CallsSet.leq env1 env2

    let join env1 env2 =
      Env.union (fun _ s1 s2 -> Some (CallsSet.join s1 s2)) env1 env2

    let meet env1 env2 =
      Env.union (fun _ s1 s2 -> Some (CallsSet.meet s1 s2)) env1 env2

    let pp = Env.pp CallsSet.pp

    let iter = Env.iter
  end

  (** The abstract domain for sets of values: the context part of
     abstract closures are elements in [CtxtEnv.t]. *)
  module Abs = AbsDomain(CtxtEnv)(Bools)(Ints)

  module Common = GlobalCommon.Make(Log)(Options)(Calls)(CtxtEnv)(Abs)
  module GEnv = Common.GEnv
  module GEnvAbs = Common.GEnvAbs
  let cofix = Common.cofix


  (** Meet of two global environments *)
  let env_meet ~narrow:_ oenv1 oenv2 = match oenv1, oenv2 with
    | None, _ | _, None -> None
    | Some env1, Some env2 ->
      let exception Inconsistent in
      try
        Some
          (GEnv.union (fun (_i, c) v1 v2 ->
               let v12 =
                 if Calls.is_max c
                 (* it is unsound to meet variables for maximal call sites *)
                 then Abs.join v1 v2
                 else Abs.meet v1 v2
               in
               if Abs.is_bot v12
               then raise Inconsistent
               else Some v12
             )
              env1 env2
          )
      with Inconsistent -> None

  (** Join of two global environments *)
  let env_join oenv1 oenv2 = match oenv1, oenv2 with
    | None, o | o, None -> o
    | Some env1, Some env2 ->
      Some
        (GEnv.merge (fun _x ov1 ov2 -> match ov1, ov2 with
             | Some v1, Some v2 -> Some (Abs.join v1 v2)
             | None, _ | _, None -> None
           )
            env1 env2
        )

  (** [backward eval genv ctxtenv t v] is a backward analyzer, that
     returns necessary informations about the variables of [t], when
     we know it forward evaluated to [v] using [eval]. It is useful to
     refine the knowledge after a branch has been taken in the forward
     analyzer. It returns a global environment that can only contain
     the variables of [t], and for which something was learned. *)
  let rec backward eval genv ctxtenv t v =
    if Abs.is_bot v
    then None
    else
      match t.Term.term with
      | Term.Var x ->
        CallsSet.join_map
          env_join
          (fun k -> Some (GEnv.(singleton k @@ get k genv)))
          (Env.find x ctxtenv) None
        (* Some (GEnv.singleton (Env.find x genv ctxtenv) v) *)
      | Term.Int n ->
        let v = Abs.(meet v (ints (Ints.singleton n))) in
        if Abs.is_bot v
        then None
        else Some GEnv.bot
      | Term.UnknownInt ->
        let v = Abs.(meet v @@ ints Ints.(interval min_int max_int)) in
        if Abs.is_bot v
        then None
        else Some GEnv.bot
      | Term.Bool b ->
        let v = Abs.(meet v (bools (Bools.singleton b))) in
        if Abs.is_bot v
        then None
        else Some GEnv.bot
      | Term.UnknownBool ->
        let v = Abs.(meet v (bools Bools.top)) in
        if Abs.is_bot v
        then None
        else Some GEnv.bot
      | Term.Unop (NEG, t) ->
        backward eval genv ctxtenv t Abs.(ints @@ Ints.Transfer.neg v.ints)
      | Term.Binop(PLUS, t1, t2) ->
        let (_g1, v1) = eval t1
        and (_g2, v2) = eval t2 in
        let env1 =
          backward eval genv ctxtenv t1 Abs.(ints @@ Ints.Transfer.minus v.ints v2.ints)
        and env2 =
          backward eval genv ctxtenv t2 Abs.(ints @@ Ints.Transfer.minus v.ints v1.ints)
        in
        env_meet ~narrow:false env1 env2
      | Term.Binop(MINUS, t1, t2) ->
        let (_g1, v1) = eval t1
        and (_g2, v2) = eval t2 in
        let env1 =
          backward eval genv ctxtenv t1 Abs.(ints @@ Ints.Transfer.plus v.ints v2.ints)
        and env2 =
          backward eval genv ctxtenv t2 Abs.(ints @@ Ints.Transfer.(minus v1.ints v.ints))
        in
        env_meet ~narrow:false env1 env2
      | Term.Binop(LE, t1, t2) ->
        let backward_when b =
        let (_g1, v1) = eval t1
        and (_g2, v2) = eval t2 in
          let (ints1, ints2) = Ints.Transfer.le v1.ints v2.ints b in
          let env1 = backward eval genv ctxtenv t1 (Abs.ints ints1)
          and env2 = backward eval genv ctxtenv t2 (Abs.ints ints2)
          in
          env_meet ~narrow:false env1 env2
        in
        let if_true =
          if Abs.mem (`Bool true) v
          then backward_when true
          else None
        and if_false =
          if Abs.mem (`Bool false) v
          then backward_when false
          else None
        in
        env_join if_true if_false
      | Term.Binop(LT, t1, t2) ->
        let backward_when b =
          let (_g1, v1) = eval t1
          and (_g2, v2) = eval t2 in
          let (ints1, ints2) = Ints.Transfer.lt v1.ints v2.ints b in
          let env1 = backward eval genv ctxtenv t1 (Abs.ints ints1)
          and env2 = backward eval genv ctxtenv t2 (Abs.ints ints2)
          in
          env_meet ~narrow:false env1 env2
        in
        let if_true =
          if Abs.mem (`Bool true) v
          then backward_when true
          else None
        and if_false =
          if Abs.mem (`Bool false) v
          then backward_when false
          else None
        in
        env_join if_true if_false
      | Term.Binop(GE, t1, t2) ->
        let backward_when b =
          let (_g1, v1) = eval t1
          and (_g2, v2) = eval t2 in
          let (ints1, ints2) = Ints.Transfer.ge v1.ints v2.ints b in
          let env1 = backward eval genv ctxtenv t1 (Abs.ints ints1)
          and env2 = backward eval genv ctxtenv t2 (Abs.ints ints2)
          in
          env_meet ~narrow:false env1 env2
        in
        let if_true =
          if Abs.mem (`Bool true) v
          then backward_when true
          else None
        and if_false =
          if Abs.mem (`Bool false) v
          then backward_when false
          else None
        in
        env_join if_true if_false
      | Term.Binop(GT, t1, t2) ->
        let backward_when b =
          let (_g1, v1) = eval t1
          and (_g2, v2) = eval t2 in
          let (ints1, ints2) = Ints.Transfer.gt v1.ints v2.ints b in
          let env1 = backward eval genv ctxtenv t1 (Abs.ints ints1)
          and env2 = backward eval genv ctxtenv t2 (Abs.ints ints2)
          in
          env_meet ~narrow:false env1 env2
        in
        let if_true =
          if Abs.mem (`Bool true) v
          then backward_when true
          else None
        and if_false =
          if Abs.mem (`Bool false) v
          then backward_when false
          else None
        in
        env_join if_true if_false
      | Term.Binop(EQ, t1, t2) ->
        let backward_when b =
          let (_g1, v1) = eval t1
          and (_g2, v2) = eval t2 in
          let (ints1, ints2) = Ints.Transfer.eq v1.ints v2.ints b in
          let env1 = backward eval genv ctxtenv t1 (Abs.ints ints1)
          and env2 = backward eval genv ctxtenv t2 (Abs.ints ints2)
          in
          env_meet ~narrow:false env1 env2
        in
        let if_true =
          if Abs.mem (`Bool true) v
          then backward_when true
          else None
        and if_false =
          if Abs.mem (`Bool false) v
          then backward_when false
          else None
        in
        env_join if_true if_false
      | Term.Binop(NEQ, t1, t2) ->
        let backward_when b =
          let (_g1, v1) = eval t1
          and (_g2, v2) = eval t2 in
          let (ints1, ints2) = Ints.Transfer.neq v1.ints v2.ints b in
          let env1 = backward eval genv ctxtenv t1 (Abs.ints ints1)
          and env2 = backward eval genv ctxtenv t2 (Abs.ints ints2)
          in
          env_meet ~narrow:false env1 env2
        in
        let if_true =
          if Abs.mem (`Bool true) v
          then backward_when true
          else None
        and if_false =
          if Abs.mem (`Bool false) v
          then backward_when false
          else None
        in
        env_join if_true if_false
      | Term.BOrElse (t1, t2) ->
        let if_true =
          if Abs.mem (`Bool true) v
          then
            let env1_t = backward eval genv ctxtenv t1 (Abs.bools @@ Bools.singleton true)
            and env1_f = backward eval genv ctxtenv t1 (Abs.bools @@ Bools.singleton false)
            and env2_t = backward eval genv ctxtenv t2 (Abs.bools @@ Bools.singleton true)
            in
            env_join env1_t (env_meet ~narrow:false env1_f env2_t)
          else
            None
        and if_false =
          if Abs.mem (`Bool false) v
          then
            let env1_f = backward eval genv ctxtenv t1 (Abs.bools @@ Bools.singleton false)
            and env2_f = backward eval genv ctxtenv t2 (Abs.bools @@ Bools.singleton false)
            in
            env_meet ~narrow:false env1_f env2_f
          else
            None
        in
        env_join if_true if_false
      | Term.BAndAlso (t1, t2) ->
        let if_true =
          if Abs.mem (`Bool true) v
          then
            let env1_t = backward eval genv ctxtenv t1 (Abs.bools @@ Bools.singleton true)
            and env2_t = backward eval genv ctxtenv t2 (Abs.bools @@ Bools.singleton true)
            in
            env_meet ~narrow:false env1_t env2_t
          else
            None
        and if_false =
          if Abs.mem (`Bool false) v
          then
            let env1_t = backward eval genv ctxtenv t1 (Abs.bools @@ Bools.singleton true)
            and env1_f = backward eval genv ctxtenv t1 (Abs.bools @@ Bools.singleton false)
            in
            let env2_f = backward eval genv ctxtenv t2 (Abs.bools @@ Bools.singleton false)
            in
            env_join env1_f (env_meet ~narrow:false env1_t env2_f)
          else
            None
        in
        env_join if_true if_false
      | Term.Binop (TIMES, _, _)
      | Term.Lam _
      | Term.FixLam _
      | Term.If _
      | Term.App _
      | Term.Let _
        -> Some GEnv.bot

  (** The abstract evaluator: given a global environment, a calling
     context and a local environment, and a term, it computes an
     abstract value for the output of the term and a new global
     environment. This new environment will be used to feed a Kleene
     iterator. We use the same widening-aware fixpoint evaluator as
     for the ∇­CFA implementation, but we actually never make use of
     widening in this particular instance. *)
  let eval genv =
    cofix @@
    fun eval _widen calls env t ->
    let eval calls env t =
      let env =
        if Options.no_env_restrict
        then env
        else Vars.restrict_env (Term.fv t) env
      in
      eval false calls env t (** never perform widening *)
    in
    match t.Term.term with
    | Term.Bool b ->
      (GEnv.bot,
       Abs.singleton (`Bool b)
      )
    | Term.Lam (xloc, x, t0) ->
      (GEnv.bot,
       let env = Vars.restrict_env (Term.fv t) env in
       Abs.singleton (`Closure (Fun.Lam (t.id, xloc, x, t0), env))
      )
    | Term.FixLam (floc, f, xloc, x, t0) ->
      (GEnv.bot,
       let env = Vars.restrict_env (Term.fv t) env in
       Abs.singleton (`Closure (Fun.FixLam (t.id, floc, f, xloc, x, t0), env))
      )
    | Term.UnknownInt ->
      (GEnv.bot,
       Abs.ints Ints.(interval min_int max_int)
      )
    | Term.UnknownBool ->
      (GEnv.bot,
       Abs.bools Bools.top
      )
    | Term.If (t1, t2, t3) ->
      let (genv1, v1) as r1 = eval calls env t1 in
      if Options.kill_unreachable && Abs.is_bot v1
      then r1
      else
        let calls' =
          if Options.branch_is_context
          then Calls.extend calls t1.id
          else calls
        in
        let ts2 =
          if Abs.mem (`Bool true) v1
          then
            let (genv2, v2) as r2 = eval calls' env t2 in
            if Options.refine_branches
            then
              let learned =
                backward (eval calls env) genv1 env
                  t1 (Abs.singleton (`Bool true))
              in
              match env_meet ~narrow:true (Some genv2) learned with
              | None -> r2
              | Some genv' -> (genv', v2)
            else r2
          else (GEnv.bot, Abs.bot)
        and ts3 =
          if Abs.mem (`Bool false) v1
          then
            let (genv3, v3) as r3 = eval calls' env t3 in
            if Options.refine_branches
            then
              let learned =
                backward (eval calls env) genv1 env
                  t1 (Abs.singleton (`Bool false))
              in
              match env_meet ~narrow:true (Some genv3) learned with
              | None -> r3
              | Some genv' -> (genv', v3)
            else r3
          else (GEnv.bot, Abs.bot)
        in
        GEnvAbs.(join (genv1, Abs.bot) @@ join ts2 ts3)
    | Term.BOrElse (t1, t2) ->
      let (genv1, v1) = eval calls env t1 in
      if Abs.mem (`Bool true) v1
      then
        if Abs.mem (`Bool false) v1
        then (* TODO: refine branch *)
          GEnvAbs.join (genv1, Abs.singleton (`Bool true)) (eval calls env t2)
        else (genv1, Abs.singleton (`Bool true))
      else
      if Abs.mem (`Bool false) v1
      then (* TODO: refine branch *)
        GEnvAbs.join (genv1, Abs.bot) (eval calls env t2)
      else (genv1, Abs.bot)
    | Term.BAndAlso (t1, t2) ->
      let (genv1, v1) = eval calls env t1 in
      if Abs.mem (`Bool false) v1
      then
        if Abs.mem (`Bool true) v1
        then (* TODO: refine branch *)
          GEnvAbs.join (genv1, Abs.singleton (`Bool false)) (eval calls env t2)
        else (genv1, Abs.singleton (`Bool false))
      else
      if Abs.mem (`Bool true) v1
      then (* TODO: refine branch *)
        GEnvAbs.join (genv1, Abs.bot) (eval calls env t2)
      else (genv1, Abs.bot)
    | Term.Var x ->
      (GEnv.bot,
       CallsSet.join_map Abs.join (fun k -> GEnv.get k genv)
         (Env.find x env) Abs.bot)
    | Term.App (t1, t2) ->
      let (genv1, v1) as r1 = eval calls env t1 in
      if Options.kill_unreachable && Abs.is_bot v1
      then r1
      else
        let (genv2, v2) as r2 = eval calls env t2 in
        if Options.kill_unreachable && Abs.is_bot v2
        then GEnvAbs.join (genv1, Abs.bot) r2
        else
          let calls' = Calls.extend calls t.Term.id in
          let genv3_v3 =
            Abs.closures_join_map
              ~join:GEnvAbs.join
              ~bot:GEnvAbs.bot
              (fun t0 env0 ->
                 match t0 with
                 | Lam (_loc, xloc, x, t') ->
                   GEnvAbs.join (GEnv.singleton (xloc, calls') v2, Abs.bot) @@
                   eval
                     calls'
                     (Env.add x (CallsSet.singleton (xloc, calls')) env0)
                     t'
                 | FixLam (_loc, floc, f, xloc, x, t') ->
                   let clos =
                     Abs.singleton
                       (`Closure (t0, (Vars.restrict_env (Fun.fv t0) env0)))
                   in
                   GEnvAbs.join (GEnv.singleton (floc, calls') clos, Abs.bot) @@
                   GEnvAbs.join (GEnv.singleton (xloc, calls') v2, Abs.bot) @@
                   eval
                     calls'
                     (Env.add x (CallsSet.singleton (xloc, calls'))
                        (Env.add f (CallsSet.singleton (floc, calls')) env0))
                     t'
              )
              v1
          in
          GEnvAbs.(join (genv1, Abs.bot) @@
                   join (genv2, Abs.bot) @@
                   genv3_v3
                  )
    | Term.Let (xloc, x, t1, t2) ->
      let (genv1, v1) as r1 = eval calls env t1 in
      if Options.kill_unreachable && Abs.is_bot v1
      then r1
      else
        let calls' =
          if Options.let_is_context
          then Calls.extend calls t.Term.id
          else calls
        in
        GEnvAbs.(join (GEnv.singleton (xloc, calls) v1, Abs.bot) @@
                 join (genv1, Abs.bot) @@
                 eval calls' (Env.add x (CallsSet.singleton (xloc, calls')) env) t2
                )
    | Term.Int n ->
      (GEnv.bot,
       Abs.singleton (`Int n)
      )
    | Term.Binop (op, t1, t2) ->
      let (genv1, v1) as r1 = eval calls env t1 in
      if Options.kill_unreachable && Abs.is_bot v1
      then r1
      else
        let (genv2, v2) = eval calls env t2 in
        let v12 =
          match op with
          | PLUS ->
            Abs.ints @@ Ints.Transfer.plus v1.Abs.ints v2.Abs.ints
          | MINUS ->
            Abs.ints @@ Ints.Transfer.minus v1.Abs.ints v2.Abs.ints
          | TIMES ->
            Abs.ints @@ Ints.Transfer.mult v1.Abs.ints v2.Abs.ints
          | LE ->
            let ((int1_t, int2_t), (int1_f, int2_f)) =
              (Ints.Transfer.le v1.Abs.ints v2.Abs.ints true,
               Ints.Transfer.le v1.Abs.ints v2.Abs.ints false)
            in
            let when_true =
              if Ints.(is_bot int1_t || is_bot int2_t)
              then Bools.bot
              else Bools.singleton true
            and when_false =
              if Ints.(is_bot int1_f || is_bot int2_f)
              then Bools.bot
              else Bools.singleton false
            in
            Abs.bools @@ Bools.join when_true when_false
          | LT ->
            let ((int1_t, int2_t), (int1_f, int2_f)) =
              (Ints.Transfer.lt v1.Abs.ints v2.Abs.ints true,
               Ints.Transfer.lt v1.Abs.ints v2.Abs.ints false)
            in
            let when_true =
              if Ints.(is_bot int1_t || is_bot int2_t)
              then Bools.bot
              else Bools.singleton true
            and when_false =
              if Ints.(is_bot int1_f || is_bot int2_f)
              then Bools.bot
              else Bools.singleton false
            in
            Abs.bools @@ Bools.join when_true when_false
          | GE ->
            let ((int1_t, int2_t), (int1_f, int2_f)) =
              (Ints.Transfer.ge v1.Abs.ints v2.Abs.ints true,
               Ints.Transfer.ge v1.Abs.ints v2.Abs.ints false)
            in
            let when_true =
              if Ints.(is_bot int1_t || is_bot int2_t)
              then Bools.bot
              else Bools.singleton true
            and when_false =
              if Ints.(is_bot int1_f || is_bot int2_f)
              then Bools.bot
              else Bools.singleton false
            in
            Abs.bools @@ Bools.join when_true when_false
          | GT ->
            let ((int1_t, int2_t), (int1_f, int2_f)) =
              (Ints.Transfer.gt v1.Abs.ints v2.Abs.ints true,
               Ints.Transfer.gt v1.Abs.ints v2.Abs.ints false)
            in
            let when_true =
              if Ints.(is_bot int1_t || is_bot int2_t)
              then Bools.bot
              else Bools.singleton true
            and when_false =
              if Ints.(is_bot int1_f || is_bot int2_f)
              then Bools.bot
              else Bools.singleton false
            in
            Abs.bools @@ Bools.join when_true when_false
          | EQ ->
            let ((int1_t, int2_t), (int1_f, int2_f)) =
              (Ints.Transfer.eq v1.Abs.ints v2.Abs.ints true,
               Ints.Transfer.eq v1.Abs.ints v2.Abs.ints false)
            in
            let when_true =
              if Ints.(is_bot int1_t || is_bot int2_t)
              then Bools.bot
              else Bools.singleton true
            and when_false =
              if Ints.(is_bot int1_f || is_bot int2_f)
              then Bools.bot
              else Bools.singleton false
            in
            Abs.bools @@ Bools.join when_true when_false
          | NEQ ->
            let ((int1_t, int2_t), (int1_f, int2_f)) =
              (Ints.Transfer.neq v1.Abs.ints v2.Abs.ints true,
               Ints.Transfer.neq v1.Abs.ints v2.Abs.ints false)
            in
            let when_true =
              if Ints.(is_bot int1_t || is_bot int2_t)
              then Bools.bot
              else Bools.singleton true
            and when_false =
              if Ints.(is_bot int1_f || is_bot int2_f)
              then Bools.bot
              else Bools.singleton false
            in
            Abs.bools @@ Bools.join when_true when_false
        in
        GEnvAbs.(join (genv1, Abs.bot) @@
                 join (genv2, Abs.bot) @@
                 (GEnv.bot, v12))
    | Term.Unop (NEG, t1) ->
      let (genv1, v1) = eval calls env t1 in
      (genv1, Abs.ints @@ Ints.Transfer.neg v1.Abs.ints)

  (** Prints statistics about the global environment on [stderr] *)
  let print_stats genv =
    if Options.debug
    then
      Log.eprintf "Bindings in global environment: %i@."
        (GEnv.cardinal genv)


  (** Checks whether there is a cycle in the global environment.
     Cycles are innocuous, but they are hints that indicate where the
     ∇-CFA analysis may be less precise. *)
  let check_genv genv =
    let finished = ref Common.BoolGEnv.empty in
    let exception Cycle in
    let exception Unbound in
    let rec browse name abs =
      let open Common.BoolGEnv in
      match find_opt name !finished with
      | None -> (* not visited yet *)
        finished := add name false !finished;
        Abs.closures_iter (fun _fun envs ->
            CtxtEnv.iter (fun _x calls_set ->
                CallsSet.iter (fun name' ->
                    match GEnv.find_opt name' genv with
                    | Some abs' -> browse name' abs'
                    | None -> raise Unbound
                  )
                  calls_set
              )
              envs)
          abs;
        finished := add name true !finished;
      | Some true -> (* visited and finished: sharing detected *)
        ()
      | Some false -> (* visited and not finished: cycle detected *)
        raise Cycle
    in
    try GEnv.iter browse genv with
    | Cycle ->
      Log.eprintf "INFO: Cycle found in global environment.@."
    | Unbound ->
      Log.eprintf "ERROR: Unbound entry in global environment. This should not happen.@."

  (** The final evaluator: computes a fixpoint of the inner evaluator *)
  let eval t =
    Common.make_eval
      ~print_stats
      ~check_genv
      eval
      t

end
