open Js_of_ocaml
open Js_of_ocaml_tyxml.Tyxml_js
open JsDemo

let ( >>= ) m f = match m with
  | Error _ as err -> err
  | Ok v -> f v

let return x = Ok x

module Options : OPTIONS = struct
  module Examples = struct
    let collection = Examples.examples
    let default = "mc91"
    let label = "Example programs:"
  end

  module Color = struct
    let normal_text = CSS.Color.(Name Black)
    let error_text = CSS.Color.(Name Red)
    let disabled_text = CSS.Color.(Name Gray)
  end

  module Button = struct
    let label = "Analyse!"
    let dom_id = "analyse-button"
    let dom_class = ["analyse-button"]
  end

  module Output = struct
    let label = "Output:"
    let dom_id = "output"
    let dom_class = ["output"]
    let font_size = 16
  end

  module Editor = struct
    let dom_id = "editor"
    let dom_class = ["editor"]
    let tab_size = 2
    let theme = "ace/theme/chrome"
    let mode = "ace/mode/ocaml"
    let font_size = 16
  end

  module OptionsDialog = struct
    type value = CfaLib.Main.Options.t

    let items = []
    let modify_onchange _ = ()
    let analysis, get_analysis, modify_onchange_analysis =
      let open CfaLib.Main.AnalysisKind in
      Utils.make_select ~name:"Analysis:" ~min ~max ~of_enum ~to_string ~default:0

    let items = analysis :: items
    let modify_onchange x =
      modify_onchange x;
      modify_onchange_analysis x

    let integer_domain, get_integer_domain, modify_onchange_integer_domain =
      let open CfaLib.Main.IntegerDomain in
      Utils.make_select ~name:"Integer domain:" ~min ~max ~of_enum ~to_string ~default:(to_enum Intervals)

    let items = integer_domain :: items
    let modify_onchange x =
      modify_onchange x;
      modify_onchange_integer_domain x

    let call_domain, get_call_domain, modify_onchange_call_domain =
      let min = 0
      and max = 5
      and of_enum = let open CfaLib.Main.CallDomain in
        function
        | 0 -> Some (LastCallSites 0)
        | 1 -> Some (LastCallSites 1)
        | 2 -> Some (LastCallSites 2)
        | 3 -> Some (CyclicCallSites 0)
        | 4 -> Some (CyclicCallSites 1)
        | 5 -> Some (CyclicCallSites 2)
        | _ -> None
      and to_string = function
        | CfaLib.Main.CallDomain.LastCallSites n ->
          Format.asprintf "last %i call sites" n
        | CfaLib.Main.CallDomain.CyclicCallSites n ->
          Format.asprintf "max %i cyclic call sites" n
      in
      Utils.make_select ~name:"Call sites approximation strategy:" ~min ~max ~of_enum ~to_string ~default:4

    let items = call_domain :: items
    let modify_onchange x =
      modify_onchange x;
      modify_onchange_call_domain x

    let refine_branches, get_refine_branches, modify_onchange_refine_branches =
      Utils.make_checkbox ~name:"Refine branches" ~checked:true
    let modify_onchange x =
      modify_onchange x;
      modify_onchange_refine_branches x

    let items = refine_branches :: items
    let branch_is_context, get_branch_is_context, modify_onchange_branch_is_context =
      Utils.make_checkbox ~name:"Branch is context" ~checked:true

    let items = branch_is_context :: items
    let modify_onchange x =
      modify_onchange x;
      modify_onchange_branch_is_context x

    let let_is_context, get_let_is_context, modify_onchange_let_is_context =
      Utils.make_checkbox ~name:"Let binding is context" ~checked:false
    let items = let_is_context :: items
    let modify_onchange x =
      modify_onchange x;
      modify_onchange_let_is_context x

    let no_env_restrict, get_no_env_restrict, modify_onchange_no_env_restrict =
      Utils.make_checkbox ~name:"No environment restriction" ~checked:false
    let items = no_env_restrict :: items
    let modify_onchange x =
      modify_onchange x;
      modify_onchange_no_env_restrict x

    let kill_unreachable, get_kill_unreachable, modify_onchange_kill_unreachable =
      Utils.make_checkbox ~name:"Kill unreachable parts" ~checked:false
    let items = kill_unreachable :: items
    let modify_onchange x =
      modify_onchange x;
      modify_onchange_kill_unreachable x

    let debug, get_debug, modify_onchange_debug =
      Utils.make_checkbox ~name:"Debug" ~checked:false
    let items = debug :: items
    let modify_onchange x =
      modify_onchange x;
      modify_onchange_debug x

    let share_integers_program_points, get_share_integers_program_points, modify_onchange_share_integers_program_points =
      Utils.make_checkbox ~name:"Share program points for identical integers" ~checked:false
    let items = share_integers_program_points :: items
    let modify_onchange x =
      modify_onchange x;
      modify_onchange_share_integers_program_points x

    let share_variables_program_points, get_share_variables_program_points, modify_onchange_share_variables_program_points =
      Utils.make_checkbox ~name:"Share program points for identical variables of the same scope" ~checked:true
    let items = share_variables_program_points :: items
    let modify_onchange x =
      modify_onchange x;
      modify_onchange_share_variables_program_points x

    let html_elt = let open Html in
      fieldset ~legend:(legend [txt "Analyser options:"])
        ~a:[a_id "options"]
        (List.rev items)

    let get_value () =
      let open CfaLib.Main.Options in
      { analysis_kind = get_analysis ();
        integer_domain = get_integer_domain ();
        call_domain = get_call_domain ();
        refine_branches = get_refine_branches ();
        branch_is_context = get_branch_is_context ();
        let_is_context = get_let_is_context ();
        no_env_restrict = get_no_env_restrict ();
        kill_unreachable = get_kill_unreachable ();
        debug = get_debug ();
        share_integers_program_points = get_share_integers_program_points ();
        share_variables_program_points = get_share_variables_program_points ();
      }

    let set_onchange = modify_onchange
  end

  let analyzer options input =
    let module F = struct
      let buf = Buffer.create 1024
      let std_formatter = Format.formatter_of_buffer buf
      let err_formatter = Format.formatter_of_buffer buf
    end in
    let module Log = Logs.Make(F) in
    let open CfaLib.Main in
    let module Analyzer =
      (val Options.get_analysis_module (module Log) options)
    in
    let run () =
      let open Lang in
      CfaLib.Driver.parse Parser.program "" (`String input) >>= fun prog ->
      Ast.Outer.check prog >>= fun () ->
      let prog =
        let open Options in
        Ast.Term.translate
          ~share_integers:options.share_integers_program_points
          ~share_variables:options.share_variables_program_points
          prog
      in
      return @@ Analyzer.eval prog
    in
    let mres, duration = Time.call run in
    Result.map_error (fun (loc, msg) ->
        Log.eprintf "%a@.%s@." Location.pp loc msg;
        Buffer.contents F.buf
      )
    @@ Result.map (fun (res, stats) ->
        Log.printf "%a@.@.(* Running time: %a *)@.(* Statistics:@.   %a *)@."
          Analyzer.Abs.pp res
          Time.pp duration
          Memo.Statistics.pp stats;
        Buffer.contents F.buf
      )
      mres

end

let () =
  let module Demo = Make(Options) in
  Demo.setup ()
