(** @author: Benoît Montagu <benoit.montagu@inria.fr>
    Copyright © Inria 2020
*)

module IntegerDomain = struct
  type t =
    | TwoPoints
    | Flat
    | Signs
    | Intervals
  [@@deriving enum]

  let parser = function
    | "two-points" -> Ok TwoPoints
    | "flat" -> Ok Flat
    | "signs" -> Ok Signs
    | "intervals" -> Ok Intervals
    | s -> Error (`Msg ("Unknown domain for integers: " ^ s))

  let to_string = function
    | TwoPoints -> "two-points"
    | Flat -> "flat"
    | Signs -> "signs"
    | Intervals -> "intervals"

  let printer fmt x =
    Format.pp_print_string fmt (to_string x)

  let examples =
    List.map (Format.asprintf "%a" printer)
      [ TwoPoints; Flat; Signs; Intervals ]

  let default = Intervals

  let get_module = let open AbstractDomains in function
    | TwoPoints -> (module TwoPoints: Sigs.INT_DOMAIN)
    | Flat -> (module FlatInts: Sigs.INT_DOMAIN)
    | Signs -> (module Signs: Sigs.INT_DOMAIN)
    | Intervals -> (module Intervals: Sigs.INT_DOMAIN)
end

module CallDomain = struct
  type t =
    | LastCallSites of int
    | CyclicCallSites of int

  let parser s =
    try
      match Scanf.sscanf s "last-call-sites:%u" (fun i -> i) with
      | i -> Ok (LastCallSites i)
      | exception Scanf.Scan_failure _ -> begin
          match Scanf.sscanf s "cyclic-call-sites:%u" (fun i -> i) with
          | i -> Ok (CyclicCallSites i)
        end
    with
    | Scanf.Scan_failure _
    | Failure _
    | End_of_file ->
      Error (`Msg ("Invalid argument: " ^ s))

  let printer fmt = function
    | LastCallSites n -> Format.fprintf fmt "last-call-sites:%i" n
    | CyclicCallSites n -> Format.fprintf fmt "cyclic-call-sites:%i" n

  let examples =
    List.map (Format.asprintf "%a" printer)
      [ CyclicCallSites(0); CyclicCallSites(1);
        LastCallSites(0); LastCallSites(1) ]

  let default = CyclicCallSites 1

  let get_module = let open Analyzers in function
    | CyclicCallSites n ->
      let module Options = struct
        let max_nesting = n
      end in
      let module M = Calls.CyclicCallSites(Options) in
      (module M: Calls.S)
    | LastCallSites n ->
      let module Options = struct
        let max_length = n
      end in
      let module M = Calls.LastCallSites(Options) in
      (module M: Calls.S)
end

module AnalysisKind = struct
  type t =
    | Nabla
    | Global
    | GlobalIndependentAttribute
  [@@deriving enum]

  let to_string = function
    | Nabla -> "nabla"
    | Global -> "global"
    | GlobalIndependentAttribute -> "global-independent-attribute"
end

module Options = struct
  type t = {
    refine_branches: bool;
    branch_is_context: bool;
    let_is_context: bool;
    no_env_restrict: bool;
    kill_unreachable: bool;
    debug: bool;
    integer_domain: IntegerDomain.t;
    call_domain: CallDomain.t;
    analysis_kind: AnalysisKind.t;
    share_integers_program_points: bool;
    share_variables_program_points: bool;
  }

  module type S = sig
    val refine_branches: bool
    val branch_is_context: bool
    val let_is_context: bool
    val no_env_restrict: bool
    val kill_unreachable: bool
    val debug: bool
  end

  let get_module options =
    let module Options = struct
      let refine_branches = options.refine_branches
      let branch_is_context = options.branch_is_context
      let let_is_context = options.let_is_context
      let no_env_restrict = options.no_env_restrict
      let kill_unreachable = options.kill_unreachable
      let debug = options.debug
    end in
    (module Options : S)

  module type ANALYZER = sig
    module Abs : sig
      type t
      val pp: Format.formatter -> t -> unit
    end
    val eval: Lang.Ast.Term.t -> (Abs.t * Memo.Statistics.t)
  end

  let get_analysis_module (logs: (module Logs.S)) options =
    let module Log = (val logs) in
    let module Options = (val get_module options) in
    let module I = (val IntegerDomain.get_module options.integer_domain) in
    let module C = (val CallDomain.get_module options.call_domain) in
    let module B = AbstractDomains.Bools in
    let open Analyzers in
    begin match options.analysis_kind with
      | Nabla ->
        (module Nabla.Analyze(Log)(Options)(C)(B)(I) : ANALYZER)
      | Global ->
        (module Global.Analyze(Log)(Options)(C)(B)(I): ANALYZER)
      | GlobalIndependentAttribute ->
        (module GlobalIndependentAttribute.Analyze(Log)(Options)(C)(B)(I) : ANALYZER)
    end
end

let ( >>= ) m f = match m with
  | Error _ as err -> err
  | Ok v -> f v

let return x = Ok x

type package =
  | Pack: 'res * (Format.formatter -> 'res -> unit) * Memo.Statistics.t -> package

type _ output_choice =
  | Value: package output_choice
  | Print: unit output_choice

let run (type a res) (out: res output_choice)
    (pp: Format.formatter -> a -> unit) : _ -> res = function
  | Error (loc, msg) -> Format.printf "%a@.%s@." Location.pp loc msg; exit 2
  | Ok (v, stats) -> begin match out with
      | Print ->
        Format.printf "@[%a@]@." pp v
      | Value ->
        Pack (v, pp, stats)
    end

let cfa_gen out options file =
  let analyze pp interpret =
    let open Lang in
    run out pp @@ begin
      Driver.parse_file Parser.program file >>= fun prog ->
      Ast.Outer.check prog >>= fun () ->
      let prog =
        let open Options in
        Ast.Term.translate
          ~share_integers:options.share_integers_program_points
          ~share_variables:options.share_variables_program_points
          prog
      in
      return @@ interpret prog
    end
  in
  let module Analyzer =
    (val Options.get_analysis_module (module Logs.Std) options)
  in
  let open Analyzer in
  analyze Abs.pp eval

let cfa_print = cfa_gen Print

let cfa_value = cfa_gen Value
