(** @author: Benoît Montagu <benoit.montagu@inria.fr>
    Copyright © Inria 2020
*)

(** Module signature for types equipped with a pretty-printer *)
module type WITH_PP = sig
  type t
  val pp: Format.formatter -> t -> unit
end

(** Module signature for parameterized types equipped with a pretty-printer *)
module type WITH_PP1 = sig
  type 'a t
  val pp: (Format.formatter -> 'a -> unit) -> Format.formatter -> 'a t -> unit
end

(** A minimalistic module signature for finite maps *)
module type MAP = sig
  type 'a t
  type key
  val empty: 'a t
  val find_opt: key -> 'a t -> 'a option
  val update: key -> ('a option -> 'a option) -> 'a t -> 'a t
  val cardinal: 'a t -> int
  include WITH_PP1 with type 'a t := 'a t
end

(** A larger module signature for finite maps *)
module type MAP2 = sig
  include MAP
  val merge: (key -> 'a option -> 'b option -> 'c option) -> 'a t -> 'b t -> 'c t
  (** [fold_similar f k m acc] folds the function [f] over the
     elements that are "similar" to the key [k] in the map [m],
     starting with accumulator [acc] *)
  val fold_similar: (key -> 'a -> 'b -> 'b) -> key -> 'a t -> 'b -> 'b
end

(** Abstract domain *)
module type DOMAIN = sig
  type t

  (** Bottom element *)
  val bot: t

  (** Pre-order *)
  val leq: t -> t -> bool

  (** Over-approximation of the upper-bound *)
  val join: t -> t -> t

  (** Over-approximation of the upper-bound, that ensures convergence:
     for any sequence [(v_n)_n], the sequence [(u_n)_n], defined as
     follows, must be stationary:
      * [u_0]  is arbitrary
      * [u_(n+1) = widen u_n v_n]
  *)
  val widen: t -> t -> t
  include WITH_PP with type t := t
end

(** Functor to create memoizing functions *)
module MakeMemo
    (Log: Logs.S)(X: WITH_PP)(M: MAP with type key = X.t)(V: WITH_PP) :
sig
  (** [memo f] memoizes the [(v, f v)] pairs in a table, so that
     subsequent calls to the memoized [f] are faster *)
  val memo: (X.t -> V.t) -> (X.t -> V.t)

  (** [fix ~transient ~debug f] memoizes the functional [f] so that
     intermediate results in the computation of the fixpoint of [f]
     are recorded in a table for faster computation. This function is
     meant to exploit memoization when defining a recursive function.
     If [transient = true], then the table persists, and is also used
     to memoize subsequent calls to the fixpoint. This means that if
     [transient = false], then the cache is not shared between
     different calls to the fixpoint (the cache is emptied at the end
     of each call). If [debug = true], then the contents of the cache
     is printed at the end of the fixpoint computation. *)
  val fix: ?transient:bool -> ?debug:bool ->
    ((X.t -> V.t) -> (X.t -> V.t)) -> (X.t -> V.t)
end

(** Statistics printed by the [MakeFix] module *)
module Statistics: sig
  type t = {
    states: int; (** number of explored states (nodes in the call graph) *)
    edges: int; (** number of explored edges (edges in the call graph) *)
    max_fanout: int; (** maximum fanout degree of states (indicates large disjunctions) *)
    iterations: int; (** number of iterations performed to reach a fixpoint *)
  }

  (** Zero-filled statistics *)
  val zero: t

  (** Pretty-printer *)
  val pp: Format.formatter -> t -> unit
end

(** Functor to compute post-fixpoints of recursively defined functions
   with values in a semi-lattice equipped with widening. The goal is
   to compute a function [f] of type [X.t -> D.t], that is a
   post-fixpoint of a functional of type [(bool -> X.t -> D.t) ->
   (bool -> X.t -> D.t)]. Both [X] and [D] must be equipped with a
   widening. The boolean argument indicates whether the fixpoint
   solver should use widening on [X]. *)
module MakeFix
    (Log: Logs.S)
    (X: sig
       include WITH_PP
       val compare: t -> t -> int
       val widen: t -> t -> t
     end)
    (M: MAP2 with type key = X.t)
    (D: DOMAIN) :
sig
  (** [cofix ~debug ~transient g] computes a function [f] such that [g
     f b x <= f b x] for every [b] and [x]. The boolean argument
     indicates whether the fixpoint solver should use widening on the
     inputs in [X] for a particular call. *)
  val cofix: ?debug:bool -> ?transient:bool ->
    ((bool -> X.t -> D.t) -> (bool -> X.t -> D.t)) ->
    (X.t -> D.t * Statistics.t)
end
